#!/usr/bin/env python3
# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
from __future__ import print_function

import argparse
import gzip
import shutil
from datetime import datetime
import glob
import redis
from redis import ConnectionError
import time
from collections import defaultdict,deque
import os, errno

__BACKUP_DIR = '/var/backup'
__DATA_DIR = '/var/data/redis/master'
__RDB = 'dump.rdb'
__AOF = 'appendonly.aof'

__DEFAULT_SAVEPOINTS = '900 1 300 10 60 10000'
__ABSOLUTE_MIN_DISK = 300 * 1024 * 1024  # Won't go less than this  disk
__MAX_FILES = 576  # 2 files every 5 mins for 24 hrs
__MAX_SLIDING_WINDOW = 7 * 24 * 60 * 60  # every 7 days

__HOST = '127.0.0.1'
__PORT = 6379


def log(*args):
    ts = datetime.now().strftime('%Y-%m-%d_%H%M%S')
    print('cts-backup',ts, *args)


def get_latest_activity():
    try:
        return redis.StrictRedis(__HOST, __PORT).info('persistence')
    except ConnectionError as ce:
        log("REDIS connection error %s" % (ce))
    except redis.exceptions.RedisError as re:
        log("REDIS error %s" % (re))


def check_minimum_disk(backup_dir, min_allowed_space=__ABSOLUTE_MIN_DISK):
    min_allowed_space = min_allowed_space if min_allowed_space > __ABSOLUTE_MIN_DISK else __ABSOLUTE_MIN_DISK
    _, _, available = shutil.disk_usage(backup_dir)

    # Removed forced AOF rewrite on theory it would topple constrained disk

    try:
        conn = redis.StrictRedis(__HOST, __PORT)
        # Pre-emptively shut off point-in-time
        if available < min_allowed_space:
            log('stopping point-in-time rdb backup - available(%s) required(%s)' % (available,min_allowed_space))
            conn.config_set('SAVE', '')
        else:
            # Check to see if someone else set savepoints
            current = conn.config_get('SAVE')
            if not current['save']:
                conn.config_set('SAVE', __DEFAULT_SAVEPOINTS)
    except ConnectionError as ce:
        log("REDIS connection error %s" % (ce))
    except redis.exceptions.RedisError as re:
        log("REDIS error %s" % (re))


def remove_backup(src):
    try:
        os.remove(src)
    except OSError as ose:
        log('could not remove backup %s %s' % (src,ose))


# Keep files:
#1. A 2 hour sliding window that tries to take snapshots every 5 minutes
#2. A 2-6 hour-old 4 hour trailing window that we keep only one snapshot as close to every 15 minute mark as possible.
#3. A 6-24 hour-old 18 hour trailing window that we keep only one snapshot per 30 minutes.
#4. 1 backup a day for 7 days prior.

_1_MIN = 60
_1_HR = 60*_1_MIN
_1_DAY = 24*_1_HR
_15_MIN = _1_MIN * 15
_30_MIN = _1_MIN * 30

_INTERVALS = [
    (8,1,_1_DAY),     # End is 1DAY ago, Start  = 1 + 7 DAYS
    (48,12,_30_MIN),  # End is 6HRS ago = 12*30MIN, Start is 24HRS = 6HRS + 18HRS = 48 30MIN
    (24,8,_15_MIN),   # End is 2HRS ago = 8*15MIN, Start is 6HRS = 2HRS + 4HRS = 24 15MIN
]

def _range(now,_start,_end,_step):
    return range(now - _start * _step, now - _end * _step, _step)

def _time_intervals():
    # Expect all the files are sorted by mtime

    now = int(time.time())
    for interval in _INTERVALS:
        for t in _range(now,*interval):
            yield t

def remove_by_time_interval(files):
    # Expect all the files are sorted by mtime

    remove = defaultdict(list)
    ti = _time_intervals()
    try:
        e = next(ti)
        for f in files:
            # if file mtime in interval add to files to remove
            while f[1] > e:
                e = next(ti)
            remove[e].append(f)
    except StopIteration:
        pass

    return remove


def cleanup_backups(backup_dir, pattern, min_allowed_space=__ABSOLUTE_MIN_DISK, sliding_window=__MAX_SLIDING_WINDOW, max_files=__MAX_FILES):
    '''
    :param backup_dir:
    :param pattern:pattern for glob to use find backups 'aof' or 'rdb'
    :param min_allowed_space:minimum_disk_space - 300M
    :param sliding_window:how old files need to be to be purged - 24 hrs
    :param max_files:max # of files 576
    :return:
    '''

    files = [(f, os.path.getmtime(f), os.path.getsize(f)) for f in glob.iglob(os.path.join(backup_dir, pattern))]
    if not files:
        log("No available files in %s matching %s to purge" % (backup_dir,pattern))
        return

    files.sort(key=lambda v: v[1])
    files.pop()  # Save the "youngest"
    files = deque(files)

    for k,v in sorted(remove_by_time_interval(files).items()):
        if len(v):
            v.pop()
        for fs in v:
            remove_backup(fs[0])
            log('removed:bytime:%s:%s' % (time.ctime(fs[1]),fs[0]))

    #If more files than max allowed # of files, remove oldest
    max_files = abs(max_files)
    while files and len(files) > max_files:
        fs = files.popleft()
        remove_backup(fs[0])
        log('removed:overlimit:%s:%s:%s' % (max_files, len(files), fs[1]))

    # Still need some space?
    min_allowed_space = min_allowed_space if min_allowed_space > __ABSOLUTE_MIN_DISK else __ABSOLUTE_MIN_DISK

    #If disk space getting low, purge oldest until above minimum threshold
    _, _, available = shutil.disk_usage(backup_dir)
    while files and available < min_allowed_space:
        fs = files.popleft()
        remove_backup(fs[0])
        log('removed:overspace:%s:%s:%s(%s)' % (min_allowed_space,(available-fs[2]), fs[0],fs[2]))
        _, _, available = shutil.disk_usage(backup_dir)


def get_mtime(file):
    try:
        return os.path.getmtime(os.path.realpath(file))
    except OSError as ose:
        log("could not get mtime for file %s" % (file))
    return 0

def backup(kind, data_path, backup_path, symlink_path):

    if not os.path.exists(data_path):
        log("no valid datafile to backup:%s" % (data_path))
        return

    last_aof_status = role = 'unknown'
    try:
        conn = redis.StrictRedis(__HOST, __PORT)
        info = conn.info()
        role = 'master' if info['role'] == 'master' else 'replica'

        last_aof_status = info['aof_last_write_status']

        if info['loading']:
            log("redis currently reloading, skipping backup %s" % (data_path))
            return

    except ConnectionError as ce:
        log("REDIS connection error %s" % (ce))
    except OSError as ose:
        log("OS error %s" % (ose))
    except redis.exceptions.RedisError as re:
        log("REDIS error %s" % (re))

    #TODO: dedup
    if kind=='aof' and last_aof_status != 'ok':
        log("no valid aof:%s:backup(%s):%s" % (role, last_aof_status, data_path))
        return

    lnm = os.path.basename(data_path) + '_' + role + '_' + 'latest.gz'
    link_mtime = get_mtime(os.path.join(symlink_path, lnm))
    data_mtime = get_mtime(data_path)

    if data_mtime > link_mtime:
        backup_file(role, data_path, backup_path, symlink_path)

def symlink(src, dst):
    try:
        os.symlink(src, dst)
    except OSError as ose:
        if ose.errno == errno.EEXIST:
            remove_backup(dst)
            try:
                os.symlink(src, dst)
            except OSError as ex:
                log('could not symlink %s %s %s' % (src,dst,ex) )
        else:
            log('could not symlink %s %s %s' % (src,dst,ose) )


def backup_file(role, data_path, backup_path, symlink_path, min_allowed_space=__ABSOLUTE_MIN_DISK):

    _, _, available = shutil.disk_usage(backup_path)
    file_sz = os.path.getsize(data_path)

    if available - file_sz < min_allowed_space:
        log("Limited diskspace(%s) - require(%s) cannot backup size(%s) %s" % (available,min_allowed_space,file_sz, data_path))
        return

    fmt = os.path.basename(data_path) + '_' + role + '_' + '%Y-%m-%d_%H%M%S.gz'
    lnm = os.path.basename(data_path) + '_' + role + '_' + 'latest.gz'

    fname = datetime.fromtimestamp(os.path.getmtime(data_path)).strftime(fmt)
    backup = os.path.join(backup_path, fname)

    with open(data_path, 'rb') as inp:
        with gzip.open(backup, 'wb') as outp:
            shutil.copyfileobj(inp, outp)

    symlink(backup, os.path.join(symlink_path, lnm))
    log("backed up %s to %s and %s" % (data_path,backup,os.path.join(symlink_path, lnm)))


def main():
    parser = argparse.ArgumentParser()
    #    parser.add_argument('--host', type=str, dest='host', help='REDIS host', default='127.0.0.1')
    #   parser.add_argument('--port', type=int, dest='port', help='REDIS port', default=6379)
    #    parser.add_argument('--rdb', action='store_true', help='Backup RDB')
    #    parser.add_argument('--aof', action='store_true', help='Backup AOF')
    #    parser.add_argument('--rdb_path', type=str, dest='rdb_path', help='dump.rdb path ',
    #                        default='/var/data/redis/master/dump.rdb')
    #    parser.add_argument('--aof_path', type=str, dest='aof_path', help='appendonly.aof path',
    #                        default='/var/data/redis/master/appendonly.aof')
    #    parser.add_argument('--backup', type=str, dest='backup_path', help='backup directory', default='/var/backups/redis')
    #    args = parser.parse_args()

    cleanup_backups(__BACKUP_DIR, 'dump.rdb*.gz')
    cleanup_backups(__BACKUP_DIR, 'appendonly.aof*.gz')
    backup('aof','/var/data/redis/master/appendonly.aof','/var/backup','/var/log/backup')
    backup('rdb','/var/data/redis/master/dump.rdb','/var/backup','/var/log/backup')
    #check_minimum_disk('/var/backup')

if __name__ == "__main__":
    main()